#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/math/distributions/binomial.hpp>
#include <boost/math/distributions/uniform.hpp>
#include "dataset.h"
#include "common.h"
#include <assert.h>
#include <iostream>
#include <sstream>
#include <stdlib.h>
#include <sys/stat.h>
#include <errno.h>

#include <vector>

//#define TRUNCATE_TIME_SERIES_LENGTH 6
//#define TRUNCATE_TIME_SERIES_N 60

using namespace boost;
using namespace std;

DataSet::DataSet()
{
}


void DataSet::randomWalks(long N, long length) {
    rand48 rng;
    uniform_real<> ur(0.0,1.0);
    variate_generator<rand48&,uniform_real<> > drawSample(rng, ur);
    
    this->N = N;
    
    for(int i = 0; i < N ; i++) {        
        originalTimeSeries[i].resize(length);
        for(int j = 0; j < length; j++) {
            originalTimeSeries[i][j] = 100.0;            
            for(int k = 0; k < j; k++)
                originalTimeSeries[i][j]+= drawSample()-0.5;
        }
        
        normalize(originalTimeSeries[i]);
        
    }
    
}

void DataSet::readFromFile(const char *src) {
    ifstream fin(src);
    
    fin.precision(10);
    
    assert(originalTimeSeries->size() == 0);
    
    cout << "Reading real timeseries collection from file " << src << "...\n";
    
    string line;
    
    int i = 0;
    
    while (std::getline(fin, line)) {
        
        if(i >= TIMESERIES_MAX) {
            FATAL("too many timeseries in this collection, you should increase TIMESERIES_MAX");
        }
        
        stringstream lineStream(line);
        
        double v;
        
        while(!lineStream.eof()) {
            lineStream >> v;
            originalTimeSeries[i].push_back(v);

        }
        
        normalize(originalTimeSeries[i]);
        cout << "TimeSeries " << i+1 <<  " has length " << originalTimeSeries[i].size() <<"\n";
        
        assert(originalTimeSeries[i].size() == originalTimeSeries[0].size());
       
        i++;
    }
    
    N = i;
    
    cout << "Done.\n";
    
    fin.close();
    
}


void DataSet::printMeanVar(vector<double> &sequence) {
    
    double sampleMean = 0;
    double sampleVariance = 0;
    
    for(unsigned int i = 0; i < sequence.size(); i++) {
        sampleMean += sequence[i];
    }
    sampleMean /= (double)sequence.size();
    
    for(unsigned int i = 0; i < sequence.size(); i++) {
        sampleVariance += pow(sequence[i]+sampleMean, 2);
    }
    
    sampleVariance = sqrt(sampleVariance/(double)sequence.size());
    
    cout << "(mean,var)=(" << sampleMean << "," << sampleVariance << ")";
}

void DataSet::normalize(vector<double> &sequence) {
    
    double sampleMean = 0;
    double sampleVariance = 0;
    
    for(unsigned int i = 0; i < sequence.size(); i++) {
        sampleMean += sequence[i];
    }
    sampleMean /= (double)sequence.size();
    
    for(unsigned int i = 0; i < sequence.size(); i++) {
        sampleVariance += pow(sequence[i]-sampleMean, 2);
    }
    
    sampleVariance = sampleVariance/(double)sequence.size();
    double sampleStdDev = sqrt(sampleVariance);
    
    if(sampleStdDev < 1e-5) {
        sampleStdDev = 1e-5;
    }
    
    for(unsigned int i = 0; i < sequence.size(); i++) {
        sequence[i] = (sequence[i] - sampleMean) / sampleStdDev;
    }
    
}


void DataSet::writeMunich(const char *dst) {

    cout << "Writing all timeseries to Munich file " << dst << " ...\n";
    
    ofstream fout(dst);
    
    if(!fout) {
        FATAL("error opening outputfile " << dst << ": " << strerror(errno));
    }
    
    
    
    for( int i = 0; i < N; i++) {
        
        fout.precision(4);
        
        for(unsigned int k = 0; k < originalTimeSeries[i].size(); k++) {


            fout << originalTimeSeries[i][k] << " : ";
            for(int j = 0; j < SAMPLES_MAX; j++) {
                fout << perturbatedTimeSeries[i][j][k] << (j == SAMPLES_MAX-1 ? "\n" : " ");
            }
        }
        
        fout << "--\n";
    }
    
    fout.close();
    cout << "Done.\n";
    
}



void DataSet::perturbateUniform(double stddev) {
    rand48 rng(static_cast<unsigned> (time(0)));
    
    double b = sqrt(3.0) * stddev;
    
    uniform_real<> ud(-b, b); // parameters: [min, max]
    variate_generator<rand48&,uniform_real<> > drawSample(rng, ud);
    
    cout << "adding Uniform perturbation (stddev=" << stddev << ") ...\n";
    
    for( int i = 0; i < N; i++) {
        cout << "processing TimeSeries " << i+1 << " ...\n";
        
        // resizing vectors
        for(unsigned int j = 0; j < SAMPLES_MAX; j++) {
            perturbatedTimeSeries[i][j].resize(originalTimeSeries[i].size());
        }
        perturbatedTimeSeriesStdDevs[i].resize(originalTimeSeries[i].size());
        
        // generating perturbated values
        for(int j = 0; j < SAMPLES_MAX; j++) {
            for(unsigned int k = 0; k < perturbatedTimeSeries[i][j].size(); k++) {
                perturbatedTimeSeries[i][j][k] = originalTimeSeries[i][k] + drawSample();
                if(j == 0) {
                    perturbatedTimeSeriesStdDevs[i][k] = stddev;
                }
            }
        }
    }
    
    //   normalizePerturbatedTimeSeries();
}



void DataSet::perturbateUNEMixed(double stddev0, double stddev1) {
    rand48 rng(static_cast<unsigned> (time(0)));

    binomial_distribution<> bd(1, 0.2);
    variate_generator<rand48&,binomial_distribution<> > drawSampleCoing(rng, bd);

    uniform_int<> une(1, 3);
    variate_generator<rand48&, uniform_int<> > whichDistrib(rng, une);



    normal_distribution<> nd0(0, stddev0);
    normal_distribution<> nd1(0, stddev1);
    variate_generator<rand48&,normal_distribution<> > nDrawSample0(rng, nd0);
    variate_generator<rand48&,normal_distribution<> > nDrawSample1(rng, nd1);


    uniform_real<> ud0(-sqrt(3.0) * stddev0, sqrt(3.0) * stddev0);
    variate_generator<rand48&,uniform_real<> > uDrawSample0(rng, ud0);
    uniform_real<> ud1(-sqrt(3.0) * stddev1, sqrt(3.0) * stddev1);
    variate_generator<rand48&,uniform_real<> > uDrawSample1(rng, ud1);


    exponential_distribution<> ed0(1.0 / stddev0);
    variate_generator<rand48&,exponential_distribution<> > eDrawSample0(rng, ed0);
    exponential_distribution<> ed1(1.0 / stddev1);
    variate_generator<rand48&,exponential_distribution<> > eDrawSample1(rng, ed1);



    cout << "adding MixedUNE perturbation (stddevs=" << stddev0 << " " << stddev1 << ") ...\n";

    for( int i = 0; i < N; i++) {

        cout << "processing TimeSeries " << i+1 << " ...\n";

        // resizing vectors
        for(unsigned int j = 0; j < SAMPLES_MAX; j++) {
            perturbatedTimeSeries[i][j].resize(originalTimeSeries[i].size());
        }
        perturbatedTimeSeriesStdDevs[i].resize(originalTimeSeries[i].size());

        int stddevId = -1;

        // generating perturbated values
        for(int j = 0; j < SAMPLES_MAX; j++) {
            for(unsigned int k = 0; k < perturbatedTimeSeries[i][j].size(); k++) {
                if(j == 0) {
                    stddevId = drawSampleCoing();
                    perturbatedTimeSeriesStdDevs[i][k] = (stddev0+stddev1)/2.0;
                }
                perturbatedTimeSeries[i][j][k] = originalTimeSeries[i][k];

                switch(whichDistrib()) {
                case 1:
                    perturbatedTimeSeries[i][j][k] += stddevId == 0 ? uDrawSample0() : uDrawSample1();
                    break;
                case 2:
                    perturbatedTimeSeries[i][j][k] += stddevId == 0 ? nDrawSample0() : nDrawSample1();
                    break;
                case 3:
                    perturbatedTimeSeries[i][j][k] += stddevId == 0 ? eDrawSample0() : eDrawSample1();
                    break;
                default:
                    assert("WTF!?" && 0);

                }
            }
        }
    }

}


void DataSet::perturbateNormalMixed(double stddev0, double stddev1) {
    rand48 rng(static_cast<unsigned> (time(0)));
    

    binomial_distribution<> bd(1, 0.2);
    variate_generator<rand48&,binomial_distribution<> > drawSampleCoing(rng, bd);
    
    normal_distribution<> nd0(0, stddev0); // parameters: mean, standard deviation.
    normal_distribution<> nd1(0, stddev1); // parameters: mean, standard deviation.
    variate_generator<rand48&,normal_distribution<> > drawSample0(rng, nd0);
    variate_generator<rand48&,normal_distribution<> > drawSample1(rng, nd1);
    
    cout << "adding MixedNormal perturbation (stddevs=" << stddev0 << " " << stddev1 << ") ...\n";
    
    for( int i = 0; i < N; i++) {
        
        cout << "processing TimeSeries " << i+1 << " ...\n";
        
        // resizing vectors
        for(unsigned int j = 0; j < SAMPLES_MAX; j++) {
            perturbatedTimeSeries[i][j].resize(originalTimeSeries[i].size());
        }
        perturbatedTimeSeriesStdDevs[i].resize(originalTimeSeries[i].size());
        
        int stddevId = -1;

        // generating perturbated values
        for(int j = 0; j < SAMPLES_MAX; j++) {
            for(unsigned int k = 0; k < perturbatedTimeSeries[i][j].size(); k++) {
                if(j == 0) {
                    stddevId = drawSampleCoing();
                    //   perturbatedTimeSeriesStdDevs[i][k] = (stddev0 + stddev1) / 2.0; // to be used for wrongxyz.
                    perturbatedTimeSeriesStdDevs[i][k] =stddevId == 0 ? stddev0 : stddev1;
                }
                perturbatedTimeSeries[i][j][k] = originalTimeSeries[i][k];
                perturbatedTimeSeries[i][j][k] += stddevId == 0 ? drawSample0() : drawSample1();
            }
        }
    }
    
}




void DataSet::perturbateExponentialMixed(double stddev0, double stddev1) {
    rand48 rng(static_cast<unsigned> (time(0)));


    binomial_distribution<> bd(1, 0.2);
    variate_generator<rand48&,binomial_distribution<> > drawSampleCoing(rng, bd);

    exponential_distribution<> ed0(1.0 / stddev0);
    exponential_distribution<> ed1(1.0 / stddev1);

    variate_generator<rand48&,exponential_distribution<> > drawSample0(rng, ed0);
    variate_generator<rand48&,exponential_distribution<> > drawSample1(rng, ed1);

    cout << "adding MixedExponential perturbation (stddevs=" << stddev0 << " " << stddev1 << ") ...\n";

    for( int i = 0; i < N; i++) {

        cout << "processing TimeSeries " << i+1 << " ...\n";

        // resizing vectors
        for(unsigned int j = 0; j < SAMPLES_MAX; j++) {
            perturbatedTimeSeries[i][j].resize(originalTimeSeries[i].size());
        }
        perturbatedTimeSeriesStdDevs[i].resize(originalTimeSeries[i].size());

        int stddevId = -1;

        // generating perturbated values
        for(int j = 0; j < SAMPLES_MAX; j++) {
            for(unsigned int k = 0; k < perturbatedTimeSeries[i][j].size(); k++) {
                if(j == 0) {
                    stddevId = drawSampleCoing();
                    //   perturbatedTimeSeriesStdDevs[i][k] = (stddev0 + stddev1) / 2.0; // to be used for wrongxyz.
                    perturbatedTimeSeriesStdDevs[i][k] =stddevId == 0 ? stddev0 : stddev1;
                }
                perturbatedTimeSeries[i][j][k] = originalTimeSeries[i][k];
                perturbatedTimeSeries[i][j][k] += stddevId == 0 ? drawSample0() : drawSample1();
            }
        }
    }

}



void DataSet::perturbateNormal(double stddev) {
    rand48 rng(static_cast<unsigned> (time(0)));
    
    normal_distribution<> nd(0, stddev); // parameters: mean, standard deviation.
    variate_generator<rand48&,normal_distribution<> > drawSample(rng, nd);
    
    cout << "adding Normal perturbation (stddev=" << stddev << ") ...\n";
    
    for( int i = 0; i < N; i++) {
        
        cout << "processing TimeSeries " << i+1 << " ...\n";
        
        // resizing vectors
        for(unsigned int j = 0; j < SAMPLES_MAX; j++) {
            perturbatedTimeSeries[i][j].resize(originalTimeSeries[i].size());
        }
        perturbatedTimeSeriesStdDevs[i].resize(originalTimeSeries[i].size());
        
        // generating perturbated values
        for(int j = 0; j < SAMPLES_MAX; j++) {
            for(unsigned int k = 0; k < perturbatedTimeSeries[i][j].size(); k++) {
                perturbatedTimeSeries[i][j][k] = originalTimeSeries[i][k] + drawSample();
                if(j == 0) {
                    perturbatedTimeSeriesStdDevs[i][k] = stddev;
                }
            }
        }
    }
    
    //  normalizePerturbatedTimeSeries();
}



void DataSet::perturbateExponential(double stddev) {
    rand48 rng(static_cast<unsigned> (time(0)));
    
    exponential_distribution<> ed(1.0 / stddev);
    variate_generator<rand48&,exponential_distribution<> > drawSample(rng, ed);
    
    cout << "adding Exponential perturbation (stddev=" << stddev << ") ...\n";
    
    for( int i = 0; i < N; i++) {
        
        cout << "processing TimeSeries " << i+1 << " ...\n";
        
        // resizing vectors
        for(unsigned int j = 0; j < SAMPLES_MAX; j++) {
            perturbatedTimeSeries[i][j].resize(originalTimeSeries[i].size());
        }
        perturbatedTimeSeriesStdDevs[i].resize(originalTimeSeries[i].size());
        
        // generating perturbated values
        for(int j = 0; j < SAMPLES_MAX; j++) {
            for(unsigned int k = 0; k < perturbatedTimeSeries[i][j].size(); k++) {
                perturbatedTimeSeries[i][j][k] = originalTimeSeries[i][k] + drawSample();
                if(j == 0) {
                    perturbatedTimeSeriesStdDevs[i][k] = stddev;
                }
            }
        }
    }
    
    //  normalizePerturbatedTimeSeries();
}



void DataSet::normalizePerturbatedTimeSeries() {
    
    return; // we do not normalizing perturbated time series, otherwise the perturbation properties change.
    
    cout << "normalizing perturbated TimeSeries ...\n";
    for(int i = 0; i < N; i++) {
        for(int j = 0; j < SAMPLES_MAX; j++) {
            normalize(perturbatedTimeSeries[i][j]);
        }
    }
}

void DataSet::writeDust(const char *dst) {
    cout << "Writing all timeseries to DUST format " << dst << " ...\n";
    
    ofstream fout(dst);
    
    if(!fout) {
        FATAL("error opening outputfile " << dst << ": " << strerror(errno));
    }
    
    fout.precision(4);
    
    cout << "saving " << N << " timeseries, length=" << originalTimeSeries[0].size() << "\n";
    
    for( int i = 0; i < N; i++) {
        
        for(unsigned int k = 0; k < originalTimeSeries[i].size(); k++) {

            fout << originalTimeSeries[i][k] << ":" << perturbatedTimeSeries[i][0][k] << ":" << perturbatedTimeSeriesStdDevs[i][k];
            
            if (k <  originalTimeSeries[i].size() -1 )
                fout << " ";
            else
                fout << "\n";
            
        }
    }
    
    fout.close();
    
    cout << "Done.\n";
}

